@@ -11,8 +11,6 @@
 #include "rpmfi-py.h"
 #include "rpmtd-py.h"
 
-#include "debug.h"
-
 /** \ingroup python
  * \class Rpm
  * \brief START HERE / RPM base module for the Python API
@@ -134,12 +132,23 @@ struct hdrObject_s {
 
 static PyObject * hdrKeyList(hdrObject * s)
 {
-    PyObject * keys = PyList_New(0);
-    HeaderIterator hi = headerInitIterator(s->h);
+    PyObject * keys;
+    HeaderIterator hi;
     rpmTagVal tag;
 
+    keys = PyList_New(0);
+    if (!keys) {
+        return NULL;
+    }
+
+    hi = headerInitIterator(s->h);
     while ((tag = headerNextTag(hi)) != RPMTAG_NOT_FOUND) {
 	PyObject *to = PyInt_FromLong(tag);
+        if (!to) {
+            headerFreeIterator(hi);
+            Py_DECREF(keys);
+            return NULL;
+        }
 	PyList_Append(keys, to);
 	Py_DECREF(to);
     }
@@ -148,36 +157,39 @@ static PyObject * hdrKeyList(hdrObject * s)
     return keys;
 }
 
-static PyObject * hdrUnload(hdrObject * s, PyObject * args, PyObject *keywords)
+static PyObject * hdrAsBytes(hdrObject * s, int legacy)
 {
-    char * buf;
-    PyObject * rc;
-    int len, legacy = 0;
-    Header h;
-    static char *kwlist[] = { "legacyHeader", NULL};
-
-    if (!PyArg_ParseTupleAndKeywords(args, keywords, "|i", kwlist, &legacy))
-	return NULL;
-
-    h = headerLink(s->h);
+    PyObject *res = NULL;
+    char *buf = NULL;
+    unsigned int len;
+    Header h = headerLink(s->h);
+   
     /* XXX this legacy switch is a hack, needs to be removed. */
     if (legacy) {
 	h = headerCopy(s->h);	/* XXX strip region tags, etc */
 	headerFree(s->h);
     }
-    len = headerSizeof(h, HEADER_MAGIC_NO);
-    buf = headerUnload(h);
+    buf = headerExport(h, &len);
     h = headerFree(h);
 
     if (buf == NULL || len == 0) {
 	PyErr_SetString(pyrpmError, "can't unload bad header\n");
-	return NULL;
+    } else {
+	res = PyBytes_FromStringAndSize(buf, len);
     }
+    free(buf);
+    return res;
+}
 
-    rc = PyBytes_FromStringAndSize(buf, len);
-    buf = _free(buf);
+static PyObject * hdrUnload(hdrObject * s, PyObject * args, PyObject *keywords)
+{
+    int legacy = 0;
+    static char *kwlist[] = { "legacyHeader", NULL};
 
-    return rc;
+    if (!PyArg_ParseTupleAndKeywords(args, keywords, "|i", kwlist, &legacy))
+	return NULL;
+
+    return hdrAsBytes(s, legacy);
 }
 
 static PyObject * hdrExpandFilelist(hdrObject * s)
@@ -220,7 +232,7 @@ static PyObject * hdrFullFilelist(hdrObject * s)
 
 static PyObject * hdrFormat(hdrObject * s, PyObject * args, PyObject * kwds)
 {
-    char * fmt;
+    const char * fmt;
     char * r;
     errmsg_t err;
     PyObject * result;
@@ -236,7 +248,7 @@ static PyObject * hdrFormat(hdrObject * s, PyObject * args, PyObject * kwds)
     }
 
     result = Py_BuildValue("s", r);
-    r = _free(r);
+    free(r);
 
     return result;
 }
@@ -294,8 +306,7 @@ static PyObject * hdrWrite(hdrObject *s, PyObject *args, PyObject *kwds)
  */
 static PyObject * hdr_fiFromHeader(PyObject * s, PyObject * args, PyObject * kwds)
 {
-    return PyObject_Call((PyObject *) &rpmfi_Type,
-			 Py_BuildValue("(O)", s), NULL);
+    return PyObject_CallFunctionObjArgs((PyObject *) &rpmfi_Type, s, NULL);
 }
 
 /* Backwards compatibility. Flags argument is just a dummy and discarded. */
@@ -308,14 +319,14 @@ static PyObject * hdr_dsFromHeader(PyObject * s, PyObject * args, PyObject * kwd
             tagNumFromPyObject, &tag, &flags))
         return NULL;
 
-    return PyObject_Call((PyObject *) &rpmds_Type,
-			 Py_BuildValue("(Oi)", s, tag), NULL);
+    return PyObject_CallFunction((PyObject *) &rpmds_Type,
+                                 "(Oi)", s, tag);
 }
 
 static PyObject * hdr_dsOfHeader(PyObject * s)
 {
-    return PyObject_Call((PyObject *) &rpmds_Type,
-			Py_BuildValue("(Oi)", s, RPMTAG_NEVR), NULL);
+    return PyObject_CallFunction((PyObject *) &rpmds_Type,
+                                 "(Oi)", s, RPMTAG_NEVR);
 }
 
 static long hdr_hash(PyObject * h)
@@ -323,6 +334,17 @@ static long hdr_hash(PyObject * h)
     return (long) h;
 }
 
+static PyObject * hdr_reduce(hdrObject *s)
+{
+    PyObject *res = NULL;
+    PyObject *blob = hdrAsBytes(s, 0);
+    if (blob) {
+	res = Py_BuildValue("O(O)", Py_TYPE(s), blob);
+	Py_DECREF(blob);
+    }
+    return res;
+}
+
 static struct PyMethodDef hdr_methods[] = {
     {"keys",		(PyCFunction) hdrKeyList,	METH_NOARGS,
 	NULL },
@@ -350,6 +372,8 @@ static struct PyMethodDef hdr_methods[] = {
 	NULL},
     {"fiFromHeader",	(PyCFunction)hdr_fiFromHeader,	METH_VARARGS|METH_KEYWORDS,
 	NULL},
+    {"__reduce__",	(PyCFunction)hdr_reduce,	METH_NOARGS,
+	NULL},
 
     {NULL,		NULL}		/* sentinel */
 };
@@ -368,8 +392,8 @@ static PyObject *hdr_new(PyTypeObject *subtype, PyObject *args, PyObject *kwds)
 
     if (obj == NULL) {
 	h = headerNew();
-    } else if (PyCObject_Check(obj)) {
-	h = PyCObject_AsVoidPtr(obj);
+    } else if (CAPSULE_CHECK(obj)) {
+	h = CAPSULE_EXTRACT(obj, "rpm._C_Header");
     } else if (hdrObject_Check(obj)) {
 	h = headerCopy(((hdrObject*) obj)->h);
     } else if (PyBytes_Check(obj)) {
@@ -457,9 +481,13 @@ static PyObject * hdrGetTag(Header h, rpmTagVal tag)
     PyObject *res = NULL;
     struct rpmtd_s td;
 
-    /* rpmtd_AsPyObj() knows how to handle empty containers and all */
     (void) headerGet(h, tag, &td, HEADERGET_EXT);
-    res = rpmtd_AsPyobj(&td);
+    if (rpmtdGetFlags(&td) & RPMTD_INVALID) {
+	PyErr_SetString(pyrpmError, "invalid header data");
+    } else {
+	/* rpmtd_AsPyobj() knows how to handle empty containers and all */
+	res = rpmtd_AsPyobj(&td);
+    }
     rpmtdFreeData(&td);
     return res;
 }
@@ -495,6 +523,8 @@ static int validData(rpmTagVal tag, rpmTagType type, rpmTagReturnType retype, Py
     } else if (retype == RPM_ARRAY_RETURN_TYPE && PyList_Check(value)) {
 	/* python lists can contain arbitrary objects, validate each item */
 	Py_ssize_t len = PyList_Size(value);
+	if (len == 0)
+	    valid = 0;
 	for (Py_ssize_t i = 0; i < len; i++) {
 	    PyObject *item = PyList_GetItem(value, i);
 	    if (!validItem(tclass, item)) {
@@ -576,7 +606,7 @@ static int hdrPutTag(Header h, rpmTagVal tag, PyObject *value)
 	    rc = hdrAppendItem(h, tag, type, item);
 	}
     } else {
-	PyErr_SetString(PyExc_RuntimeError, "cant happen, right?");
+	PyErr_SetString(PyExc_RuntimeError, "can't happen, right?");
     }
 
     return rc;
@@ -608,10 +638,18 @@ static PyObject * hdr_getattro(hdrObject * s, PyObject * n)
 {
     PyObject *res = PyObject_GenericGetAttr((PyObject *) s, n);
     if (res == NULL) {
+	PyObject *type, *value, *traceback;
 	rpmTagVal tag;
+
+	/* Save and restore original exception if it's not a valid tag either */
+	PyErr_Fetch(&type, &value, &traceback);
 	if (tagNumFromPyObject(n, &tag)) {
-	    PyErr_Clear();
+	    Py_XDECREF(type);
+	    Py_XDECREF(value);
+	    Py_XDECREF(traceback);
 	    res = hdrGetTag(s->h, tag);
+	} else {
+	    PyErr_Restore(type, value, traceback);
 	}
     }
     return res;
@@ -825,7 +863,7 @@ static int compare_values(const char *str1, const char *str2)
 
 PyObject * labelCompare (PyObject * self, PyObject * args)
 {
-    char *v1, *r1, *v2, *r2;
+    const char *v1, *r1, *v2, *r2;
     const char *e1, *e2;
     int rc;
 
